import errno
import os
import signal
import subprocess
from multiprocessing import Queue, Process, Manager
from multiprocessing.pool import Pool
from subprocess import TimeoutExpired
from typing import Optional, Union, Any

from .observe import Observer, dispatch


class Update:
    SENTINEL = "sentinel"
    STARTED = "started"
    DONE = "done"
    TIMEOUT = "timeout"

    def __init__(self, status, index, command, meta):
        self.status = status
        self.index = index
        self.command = command
        self.meta = meta


class ParallelObserver(Observer):
    @dispatch
    def observe(self, update):
        # type: (Update) -> None
        raise NotImplementedError()


def observe(observer, queue, count=None):
    # type: (ParallelObserver, Queue, Optional[int]) -> None
    to_see = None if count is None else set(range(count))
    while to_see is None or len(to_see) > 0:
        update = queue.get()  # type: Union[Update, str]
        if update == Update.SENTINEL:
            return
        elif isinstance(update, Update):
            if update.status == Update.DONE or update.status == Update.TIMEOUT:
                if to_see:
                    to_see.remove(update.index)
            observer.observe(update)
        else:
            raise ValueError("Invalid update {}".format(update))


def run_command(args):
    i, meta, command, timeout, queue = args  # type: (int, Any, str, int, Queue)
    with subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, preexec_fn=os.setsid) as process:
        try:
            if queue:
                queue.put(Update(Update.STARTED, i, command, meta))
            process.communicate(timeout=timeout)
            if queue:
                queue.put(Update(Update.DONE, i, command, meta))
        except TimeoutExpired:
            if queue:
                queue.put(Update(Update.TIMEOUT, i, command, meta))

            try:
                os.killpg(process.pid, signal.SIGINT)  # send signal to the process group
            except OSError as e:
                if e.errno != errno.ESRCH:
                    if e.errno == errno.EPERM:
                        os.waitpid(-process.pid, 0)
                else:
                    raise e

            process.communicate()


def run_commands(commands, processes=None, timeout=None, meta=None, observer=None):
    pool = Pool(processes=processes)
    manager, queue = None, None
    if observer:
        manager = Manager()
        queue = manager.Queue()

    if meta:
        commands = [(i, meta, command, timeout, queue) for i, (command, meta) in enumerate(zip(commands, meta))]
    else:
        commands = [(i, meta, command, timeout, queue) for i, command in enumerate(commands)]

    observer_process = None
    if observer:
        observer_process = Process(target=observe, args=(observer, queue, len(commands)))
        observer_process.daemon = True
        observer_process.start()

    pool.map(run_command, commands)

    if observer_process:
        observer_process.join()



import inspect
from typing import TYPE_CHECKING, Type

from pebble import ProcessPool

from .observe import ProgressObserver
from .parallel import ParallelObserver, Update
from . import parallel
from .storage import export_storage
from concurrent.futures import TimeoutError as OtherTimeoutError


class ParallelToProcess(ParallelObserver):
    def __init__(self, observer, runner):
        # type: (ProgressObserver, CommandLineRunner) -> None
        super().__init__()
        self.observer = observer
        self.runner = runner

    def observe(self, update):
        if self.observer.auto_load:
            meta = self.runner.trajectory.experiments[update.index]
            if update.status == Update.DONE:
                meta = meta.fresh_copy()
        else:
            meta = update.meta

        if update.status == Update.STARTED:
            self.observer.experiment_started(update.index, meta)
        if update.status == Update.DONE:
            self.observer.experiment_finished(update.index, meta)
        if update.status == Update.TIMEOUT:
            self.observer.experiment_interrupted(update.index, meta)


class PrintObserver(ProgressObserver):
    def experiment_started(self, index, experiment):
        print("[{}] started: {}".format(index, experiment))

    def experiment_finished(self, index, experiment):
        print("[{}] done: {}".format(index, experiment))

    def experiment_interrupted(self, index, experiment):
        print("[{}] timed out: {}".format(index, experiment))


class CommandLineRunner(object):
    def __init__(self, trajectory, storage, processes=None, timeout=None, observer=None):
        self.trajectory = trajectory
        self.storage = storage
        self.timeout = timeout
        self.processes = processes
        self.observer = None if observer is None else ParallelToProcess(observer, self)

    def run(self):
        commands = []
        for experiment in self.trajectory.experiments:
            if self.timeout:
                experiment.config["timeout"] = self.timeout
            experiment.config["started"] = True
            experiment.save(self.storage)
            storage_name = export_storage(experiment.storage)
            filename = inspect.getfile(experiment.__class__)
            commands.append("python {} {} run {}".format(filename, storage_name, experiment.identifier))
        meta = [e.identifier for e in self.trajectory.experiments] if self.observer else None
        parallel.run_commands(commands, timeout=self.timeout, observer=self.observer, meta=meta)
        return [self.storage.get_experiment(e.__class__, e.identifier) for e in self.trajectory.experiments]


class ParallelRunner(object):
    def __init__(self, trajectory, storage, processes=None, timeout=None):
        self.trajectory = trajectory
        self.storage = storage
        self.timeout = timeout
        self.processes = processes

    def run(self):
        for e in self.trajectory.experiments:
            if self.timeout:
                e.config["timeout"] = self.timeout
            e.save(self.storage)
        args_list = [(self.storage, e.__class__, e.identifier) for e in self.trajectory.experiments]
        results = []
        with ProcessPool() as pool:
            future = pool.map(ParallelRunner.run_single, args_list, timeout=self.timeout)

            iterator = future.result()

            while True:
                exp_id = None
                try:
                    exp_id = next(iterator)
                    results.append(exp_id)
                    print("[done]    {exp_id}".format(exp_id=exp_id))
                except StopIteration:
                    break
                except (TimeoutError, OtherTimeoutError):
                    print("[timeout] {exp_id} (timeout={timeout})".format(exp_id=exp_id or "-", timeout=self.timeout))

        return [self.storage.get_experiment(oe.__class__, exp_id)
                for exp_id, oe in zip(results, self.trajectory.experiments)]

    @staticmethod
    def run_single(arg_tuple):
        storage, cls, identifier = arg_tuple
        experiment = storage.get_experiment(cls, identifier)
        experiment.run(True)
        return experiment.identifier


def import_runner(runner_string, trajectory, storage, timeout=None):
    if runner_string == "cli":
        return CommandLineRunner(trajectory, storage, timeout=timeout)
    elif runner_string == "multi":
        return ParallelRunner(trajectory, storage, timeout=timeout)
    else:
        raise ValueError("Could not parse runner from {}".format(runner_string))



import os
from functools import partial

from peewee import Model, SqliteDatabase, CharField, BooleanField
from playhouse.fields import PickleField

from .storage import Storage

database = SqliteDatabase(
    os.environ.get("DB", os.path.join(os.path.dirname(os.path.dirname(__file__)), "experiments.sqlite"))
)


class BaseModel(Model):
    class Meta:
        database = database


class ExperimentModel(BaseModel):
    cls_name = CharField()
    completed = BooleanField()
    group = CharField()
    config = PickleField()
    parameters = PickleField()
    result = PickleField()
    derived = PickleField()


def class_name(cls):
    return cls.__name__


class SqliteStorage(Storage):
    def __init__(self):
        database.connect()
        database.create_tables([ExperimentModel], safe=True)
        database.close()

    def save(self, experiment):
        if experiment.storage == self and experiment.identifier:
            model = ExperimentModel.get_by_id(experiment.identifier)
            model.completed = experiment.completed
            model.group = experiment.group
            model.config = experiment.config.values
            model.parameters = experiment.parameters.values
            model.result = experiment.result.values
            model.derived = experiment.derived
            model.save()

        elif (not experiment.storage or experiment == self) and not experiment.identifier:
            cls = class_name(experiment.__class__)
            model = ExperimentModel.create(cls_name=cls, group=experiment.group, config=experiment.config.values,
                                           parameters=experiment.parameters.values, result=experiment.result.values,
                                           derived=experiment.derived, completed=experiment.completed)
            experiment.storage = self
            experiment.identifier = model.id
        else:
            if experiment != self:
                raise ValueError("Experiment comes from a different storage")
            else:
                raise ValueError("Experiment is partially instantiated")

    def transform(self, cls, model):
        experiment = cls(model.group, self, identifier=model.id)
        experiment.completed = model.completed
        for key, value in model.config.items():
            experiment.config[key] = value
        for key, value in model.parameters.items():
            experiment.parameters[key] = value
        for key, value in model.result.items():
            experiment.result[key] = value
        for key, value in model.derived.items():
            experiment.derived[key] = value
        return experiment

    def get_experiment(self, cls, identifier):
        model = ExperimentModel.get_by_id(identifier)
        if class_name(cls) == model.cls_name:
            return self.transform(cls, model)
        else:
            raise ValueError("Could not find experiment with id {} and class {} (was {})"
                             .format(identifier, cls, model.cls_name))

    def get_experiments(self, cls, group=None):
        cls_name = class_name(cls)
        if group:
            return list(map(partial(self.transform, cls),
                            ExperimentModel.select()
                            .where(ExperimentModel.cls_name == cls_name, ExperimentModel.group == group)))
        else:
            return list(map(partial(self.transform, cls),
                            ExperimentModel.select().where(ExperimentModel.cls_name == cls_name)))

    def remove(self, group):
        ExperimentModel.delete().where(ExperimentModel.group == group).execute()

    def get_groups(self):
        return list(ExperimentModel.select(ExperimentModel.group))



import itertools
import platform
import numpy

import matplotlib as mpl
import matplotlib.markers as mark
import matplotlib.pyplot as plt
import matplotlib.cm as cm

if platform.system() == "Darwin":
    mpl.use('TkAgg')


class ScatterData:
    colors = []
    markers = ["o", "v", "x"]

    def __init__(self, title, plot_options):
        self.title = title
        self.data = []
        self.limits = None, None
        self.plot_options = plot_options

    def add_data(self, name, x_data, y_data, error=None):
        self.data.append((name, x_data, y_data, error))
        return self

    @property
    def size(self):
        return len(self.data)

    def x_lim(self, limits):
        self.limits = limits, self.limits[1]

    def y_lim(self, limits):
        self.limits = self.limits[0], limits

    def gen_colors(self):
        if len(self.data) <= len(self.colors):
            return self.colors[:len(self.data)]
        iterator = iter(cm.get_cmap("rainbow")(numpy.linspace(0, 1, len(self.data))))
        return [next(iterator) for _ in range(len(self.data))]

    def gen_markers(self):
        if len(self.data) <= len(self.markers):
            return self.markers[:len(self.data)]
        iterator = itertools.cycle(mark.MarkerStyle.filled_markers)
        return [next(iterator) for _ in range(len(self.data))]

    def render(self, ax, lines=True, log_x=True, log_y=True, label_x=None, label_y=None, legend_pos=None,
               x_ticks=None, y_ticks=None):

        plots = []
        colors = self.gen_colors()
        markers = self.gen_markers()

        if legend_pos is None:
            legend_pos = "lower right"

        plot_diagonal = False
        plot_extra = None
        plot_format = "scatter"
        show_error = True
        steps_x = None
        steps_y = None

        cache = None
        for plot_option in self.plot_options or ():
            if cache is None:
                if plot_option == "diagonal":
                    plot_diagonal = True
                else:
                    cache = plot_option
            else:
                if cache == "format":
                    plot_format = plot_option
                elif cache == "error":
                    show_error = int(plot_option)
                elif cache == "legend":
                    if plot_option != "none":
                        legend_pos = plot_option
                elif cache == "lx":
                    label_x = plot_option
                elif cache == "ly":
                    label_y = plot_option
                elif cache == "steps_x":
                    steps_x = int(plot_option)
                elif cache == "steps_y":
                    steps_y = int(plot_option)
                elif cache == "plot_extra":
                    plot_extra = plot_option
                elif cache == "x_lim":
                    parts = plot_option.split(":")
                    limits = (float(parts[0]), float(parts[1]))
                    self.x_lim(limits)
                elif cache == "y_lim":
                    parts = plot_option.split(":")
                    limits = (float(parts[0]), float(parts[1]))
                    self.y_lim(limits)
                cache = None

        min_x, max_x, min_y, max_y = numpy.infty, -numpy.infty, numpy.infty, -numpy.infty
        for i in range(self.size):
            name, x_data, y_data, error = self.data[i]
            try:
                min_x = min(min_x, numpy.min(x_data))
                min_y = min(min_y, numpy.min(y_data))
                max_x = max(max_x, numpy.max(x_data))
                max_y = max(max_y, numpy.max(y_data))
            except TypeError:
                pass

            if plot_format == "scatter":
                plots.append(ax.scatter(x_data, y_data, color=colors[i], marker=markers[i], s=40))
                if lines:
                    ax.plot(x_data, y_data, color=colors[i])
                if show_error == 1 and error is not None:
                    ax.fill_between(x_data, y_data - error, y_data + error, color=colors[i], alpha=0.35, linewidth=0)
                elif show_error == 2 and error is not None:
                    ax.errorbar(x_data, y_data, error, linestyle='None', color=colors[i])
            elif plot_format == "bar":
                plots.append(ax.bar(x_data, y_data, color=colors[i]))
            else:
                raise ValueError("Unknown plot format")

            if i == self.size - 1:
                if plot_extra and plot_extra == "1/x":
                    ax.plot(x_data, 1 / x_data, linestyle="--")

        if plot_diagonal:
            ax.plot(numpy.array([min_x, max_x]), numpy.array([min_y, max_y]), linestyle="--")

        ax.grid(True)
        legend_names = list(t[0] for t in self.data)
        # legend_names = ["No mixing - DT", "No mixing - RF", "Mixing - DT", "Mixing - RF"]
        # legend_names = ["No formulas", "Formulas"]
        # legend_names = []
        if 10 > len(self.data) == len(legend_names) and legend_pos:
            ax.legend(plots, legend_names, loc=legend_pos)

        if log_x:
            ax.set_xscale('log')
        if log_y:
            ax.set_yscale('log')

        x_lim, y_lim = self.limits
        if x_lim is not None:
            ax.set_xlim(x_lim)
        if y_lim is not None:
            ax.set_ylim(y_lim)

        if label_y is not None:
            ax.set_ylabel(label_y)
        if label_x is not None:
            ax.set_xlabel(label_x)

        if steps_x is not None:
            x_ticks = numpy.linspace(min_x, max_x, steps_x)
        if steps_y is not None:
            y_ticks = numpy.linspace(min_y, max_y, steps_y)
        # x_ticks = [1, 2, 3]
        if x_ticks is not None:
            ax.xaxis.set_ticks(x_ticks)
        if y_ticks is not None:
            ax.yaxis.set_ticks(y_ticks)

    def plot(self, filename=None, size=None, **kwargs):
        fig = plt.figure()
        if size is not None:
            fig.set_size_inches(*size)
        self.render(fig.gca(), **kwargs)
        if filename is None:
            plt.show(block=True)
        else:
            plt.savefig(filename, format="png", bbox_inches="tight", pad_inches=0.08, dpi=600)



import importlib
from typing import List, TYPE_CHECKING, Optional, Type

if TYPE_CHECKING:
    from .experiment import Experiment


class Storage(object):
    def save(self, experiment):
        raise NotImplementedError()

    def get_experiment(self, cls, identifier):
        # type: (Type, int) -> Experiment
        raise NotImplementedError()

    def get_experiments(self, cls, group=None):
        # type: (Type, Optional[str]) -> List[Experiment]
        raise NotImplementedError()

    def remove(self, group):
        raise NotImplementedError()

    def get_groups(self):
        # type: () -> List[str]
        raise NotImplementedError()


def export_storage(storage):
    from .sql_storage import SqliteStorage
    if isinstance(storage, SqliteStorage):
        return "sqlite"
    else:
        raise ValueError("Could not export storage {storage}".format(storage=storage))


def import_storage(storage_string):
    if storage_string == "sqlite":
        from .sql_storage import SqliteStorage
        return SqliteStorage()
    else:
        raise ValueError("Could not import storage {storage_string}".format(storage_string=storage_string))


def full_class_name(cls):
    # Inspired by https://stackoverflow.com/a/2020083
    module = cls.__module__
    if module is None or module == str.__class__.__module__:
        return cls.__name__  # Avoid reporting __builtin__
    else:
        return module + '.' + cls.__name__


def str_to_class(n):
    # Inspired by https://stackoverflow.com/a/13808375
    parts = n.split(".")
    module_name = ".".join(parts[:-1])
    class_name = parts[-1]

    m = importlib.import_module(module_name)
    return getattr(m, class_name)



import time
from typing import Union, Any, Dict, List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from storage import Storage

from .trajectory import Trajectory


class Parameter(object):
    def __init__(self, p_type, default, description=None, name=None, arg_name=None):
        self.name = name
        self.p_type = p_type
        self.default = default
        self.description = description
        self.specific_arg_name = arg_name

    @property
    def arg_name(self):
        return self.specific_arg_name or self.name


class Result(Parameter):
    pass


class Config(Parameter):
    pass


class Group(object):
    def __init__(self, name):
        self.name = name
        self.parameters = dict()  # type: Dict[str, Parameter]
        self.values = dict()  # type: Dict[str, Any]

    def add_parameter(self, name, p_type, default=None, description=None, arg_name=None):
        self.add(Parameter(p_type, default, description, name, arg_name))

    def add(self, parameter):
        self.parameters[parameter.name] = parameter

    def set_values(self, **kwargs):
        for key, value in kwargs.items():
            self.set_value(key, value)

    def set_value(self, name, value):
        if name not in self.parameters:
            raise ValueError("No parameter called {name}".format(name=name))
        if not isinstance(value, self.parameters[name].p_type):
            raise ValueError("Unexpected type for value {value}".format(value=value))
        self.values[name] = value

    def __setitem__(self, key, value):
        self.set_value(key, value)

    def __getitem__(self, item):
        return self.values[item] if item in self.values else self.parameters[item].default

    def add_arguments(self, parser, prefix=None):
        for name, parameter in self.parameters.items():
            parser.add_argument(
                "--{}{}".format(prefix or '', parameter.arg_name),
                type=parameter.p_type,
                default=parameter.default,
                help=parameter.description
            )

    def parse_arguments(self, args, prefix=None):
        for name, parameter in self.parameters.items():
            self.set_value(name, getattr(args, "{}{}".format(prefix or '', parameter.arg_name)))

    def get_arguments(self, prefix=None):
        return " ".join(
            "--{} {}".format("{}{}".format(prefix or '', self.parameters[name].arg_name), value)
            for name, value in self.values.items() if value is not None
        )

    def copy(self):
        group = self.__new__(self.__class__)
        group.parameters = dict(self.parameters)
        group.values = dict(self.values)
        return group

    def __str__(self):
        values = ", ".join("{}: {}".format(key, self.values.get(key, '')) for key in self.parameters)
        return "{}={{{}}}".format(self.name, values)


class Derived(object):
    def __init__(self, callback):
        self.callback = callback

    def __call__(self, *args, **kwargs):
        return self.callback(*args, **kwargs)


def derived(func):
    return Derived(func)


class Experiment(object):
    def __init__(self, group, storage=None, identifier=None):
        # type: (str, Optional[Storage], Optional[int]) -> None
        self.derived_callbacks = dict()
        self.storage = storage
        self.identifier = identifier
        self.completed = False

        self.group = group
        self.config = Group("config")
        self.config.add_parameter("timeout", int, None, description="The timeout value set for this experiment")
        self.config.add_parameter("started", bool, False, description="Whether this experiment has been started")
        self.parameters = Group("parameters")
        self.result = Group("result")
        self.result.add_parameter("runtime", float, None, description="How long the experiment took to execute")
        self.derived = dict()

        for key, value in self.__class__.__dict__.items():
            if isinstance(value, Parameter):
                if value.name is None:
                    value.name = key
                if isinstance(value, Config):
                    self.config.add(value)
                elif isinstance(value, Result):
                    self.result.add(value)
                else:
                    self.parameters.add(value)
            elif isinstance(value, Derived):
                if key.startswith("derived_"):
                    key = key[8:]
                self.derived_callbacks[key] = value

    def get_derived(self, name):
        if name in self.derived:
            return self.derived[name]

        if name in self.derived_callbacks:
            result = self.derived_callbacks[name](self)
            self.derived[name] = result
            return result

        raise ValueError("There is no derived attribute with the name {name}".format(name=name))

    def __getitem__(self, item):
        return self.get(item)

    def __setitem__(self, key, value):
        self.set(key, value)

    def get(self, name: Union[str, Parameter]):
        if isinstance(name, Parameter):
            name = name.name
        parts = name.split(".", 1)
        if parts[0] == "par" or parts[0] == "parameter":
            return self.parameters[parts[1]]
        elif parts[0] == "res" or parts[0] == "result":
            return self.result[parts[1]]
        elif parts[0] == "conf" or parts[0] == "config":
            return self.config[parts[1]]
        elif parts[0] == "derived":
            return self.get_derived(parts[1])
        else:
            name = ".".join(parts)
            results = []
            if name in self.config.parameters:
                results.append(self.config[name])
            elif name in self.parameters.parameters:
                results.append(self.parameters[name])
            elif name in self.result.parameters:
                results.append(self.result[name])
            elif name in self.derived_callbacks:
                results.append(self.get_derived(name))
            if len(results) == 1:
                return results[0]
            elif len(results) > 1:
                raise ValueError("Multiple entries found for the name {name}, please use parameter.{name}, "
                                 "result.{name}, config.{name} or derived.{name} to disambiguate".format(name=name))
        raise ValueError("No entry found for the name {name}".format(name=name))

    def set(self, name, value):
        if isinstance(name, Parameter):
            name = name.name

        parts = name.split(".", 1)
        if parts[0] == "par" or parts[0] == "parameter":
            self.parameters[parts[1]] = value
        elif parts[0] == "res" or parts[0] == "result":
            self.result[parts[1]] = value
        elif parts[0] == "conf" or parts[0] == "config":
            self.config[parts[1]] = value
        elif parts[0] == "derived":
            self.get_derived(parts[1])
        else:
            name = ".".join(parts)
            results = []
            if name in self.config.parameters:
                results.append(self.config)
            elif name in self.parameters.parameters:
                results.append(self.parameters)
            elif name in self.result.parameters:
                results.append(self.result)

            if len(results) == 1:
                results[0][name] = value
            elif len(results) > 1:
                raise ValueError("Multiple entries found for the name {name}, please use parameter.{name}, "
                                 "result.{name}, config.{name} or derived.{name} to disambiguate".format(name=name))
            else:
                raise ValueError("No entry found for the name {name}".format(name=name))

    def run(self, auto_save=True):
        try:
            start = time.perf_counter()
            self.run_internal()
            runtime = time.perf_counter() - start
            self.result["runtime"] = runtime
            self.completed = True
            if auto_save and self.storage:
                self.save()
            return self
        except KeyboardInterrupt:
            return self

    def run_internal(self):
        raise NotImplementedError()

    def save(self, storage=None):
        if storage:
            storage.save(self)
        else:
            if not self.storage:
                raise ValueError("No storage specified")
            self.storage.save(self)

    def fresh_copy(self):
        return self.storage.get_experiment(self.__class__, self.identifier)

    def __str__(self):
        return "EXP(group={}, id={}, {}, {}, {}, derived={})"\
            .format(self.group or '', self.identifier or '', self.config, self.parameters, self.result, self.derived)

    def add_arguments(self, parser, prefix=None):
        self.config.add_arguments(parser, "{}conf.".format(prefix or ''))
        self.parameters.add_arguments(parser, "{}par.".format(prefix or ''))
        self.result.add_arguments(parser, "{}res.".format(prefix or ''))

    def parse_arguments(self, args, prefix=None):
        self.config.parse_arguments(args, "{}conf.".format(prefix or ''))
        self.parameters.parse_arguments(args, "{}par.".format(prefix or ''))
        self.result.parse_arguments(args, "{}res.".format(prefix or ''))

    def get_arguments(self, prefix=None):
        strings = [
            self.config.get_arguments("{}conf.".format(prefix or '')),
            self.parameters.get_arguments("{}par.".format(prefix or '')),
            self.result.get_arguments("{}res.".format(prefix or '')),
        ]
        return " ".join(strings)

    @classmethod
    def explore(cls, name: str, settings: Union[List[Dict[str, Any]], Dict[str, List]]):
        trajectory = Trajectory(name)
        trajectory.explore(cls, settings)
        return trajectory

    @classmethod
    def enable_cli(cls):
        if cls.__module__ == "__main__":
            cls.run_cli()

    @classmethod
    def run_cli(cls):
        from .cli import parse_cli
        parse_cli(cls)



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .experiment import Experiment


class Observer:
    def __init__(self):
        self.observers = []

    def add_observer(self, observer):
        self.observers.append(observer)

    def remove_observer(self, observer):
        self.observers.remove(observer)


def dispatch(func):
    def modified(self, *args, **kwargs):
        for observer in self.observers:
            getattr(observer, func.__name__)(*args, **kwargs)
    return modified


class ProgressObserver(Observer):
    def __init__(self, auto_load=True):
        super().__init__()
        self.auto_load = auto_load

    @dispatch
    def experiment_started(self, index, experiment):
        # type: (int, Experiment) -> None
        raise NotImplementedError()

    @dispatch
    def experiment_finished(self, index, experiment):
        # type: (int, Experiment) -> None
        raise NotImplementedError()

    @dispatch
    def experiment_interrupted(self, index, experiment):
        # type: (int, Experiment) -> None
        raise NotImplementedError()






import collections
from argparse import ArgumentParser
from typing import List, Optional, Union, Any

import numpy as np

from .experiment import Experiment


def table(data, column_header=None, row_header=None):
    # type: (Union[List[List[Any]], np.ndarray], Optional[List[Any]], Optional[List[Any]]) -> str

    if isinstance(data, np.ndarray):
        # noinspection PyTypeChecker
        data = [[data[row, col] for col in range(data.shape[1])] for row in range(data.shape[0])]

    if row_header is not None:
        if not isinstance(row_header[0], (list, tuple)):
            row_header = [row_header]
        for rh in reversed(row_header):
            row_strings = ["{}".format(rh[i]) for i in range(len(data))]
            max_length = max(len(rs) for rs in row_strings)
            data = [[row_strings[i]] + data[i] for i in range(len(data))]
            if column_header is not None:
                column_header = [" " * (max_length - 1) + " "] + list(column_header)

    if column_header is not None:
        column_header = ["{}".format(e) for e in column_header]
    data = [["{}".format(e) for e in row] for row in data]

    max_lengths = [max(len(data[r][c]) for r in range(len(data))) for c in range(len(data[0]))]
    if column_header is not None:
        max_lengths = [max(max_lengths[c], len(column_header[c])) for c in range(len(data[0]))]

    if column_header is not None:
        data.insert(0, column_header)
        data.insert(1, ["=" * max_lengths[i] for i, c in enumerate(column_header)])

    data = [[row[c] + " " * (max_lengths[c] - len(row[c])) for c in range(len(data[0]))] for row in data]

    return "\n".join("  ".join(e for e in row) for row in data)


def add_arguments(parser: ArgumentParser):
    parser.add_argument("-t", "--targets", nargs="+", type=str, default=None)
    parser.add_argument("-g", "--group_by", nargs="+", type=str, default=None)
    parser.add_argument("-a", "--aggregator", type=str, default=None)
    parser.add_argument("-s", "--sort", type=str, default=None)
    parser.add_argument("-e", "--exclude", nargs="+", type=str, default=None)
    parser.add_argument("-p", "--plot", action="store_true")
    parser.add_argument("-o", "--options", nargs="+", type=str, default=None)
    parser.add_argument("-w", "--write_to", type=str, default=None)


def mean(iterable):
    return sum(iterable) / len(iterable)


def mean_scores(scores):
    return [mean([scores[r][c] for r in range(len(scores))]) for c in range(len(scores[0]))]


def std_scores(scores):
    return [np.std(np.array([scores[r][c] for r in range(len(scores))])) for c in range(len(scores[0]))]


def parse_args(args):
    return args.targets, args.group_by, args.aggregator, args.sort, args.exclude, args.plot, args.options, args.write_to


def show_from_args(experiments, args):
    show(experiments, *parse_args(args))


def get_property(index: int, experiment: Experiment, property_name: str):
    parts = property_name.split("__")
    if len(parts) > 1:
        operator = parts[-1]
        remaining = "__".join(parts[:-1])
        if operator == "mean":
            return mean(get_property(index, experiment, remaining))
        if operator == "len":
            return len(get_property(index, experiment, remaining))
        if operator == "sum":
            return sum(get_property(index, experiment, remaining))
        if operator == "max":
            return max(get_property(index, experiment, remaining))
        if operator == "min":
            return min(get_property(index, experiment, remaining))
        if operator == "test":
            return 1 if is_excluded_from_string(remaining, experiment) else 0
        if operator.startswith("batch"):
            bin_size = float(operator[5:])
            return int(get_property(index, experiment, remaining) / bin_size) * bin_size - bin_size / 2
        else:
            try:
                index = int(operator)
                return get_property(index, experiment, remaining)[index]
            except ValueError:
                pass

    if property_name == "id":
        return index

    if property_name == "*":
        return "all"

    if property_name == "group":
        return experiment.group

    if property_name == "completed":
        return experiment.completed

    try:
        return experiment.get(property_name)
    except ValueError:
        pass

    if "." in property_name:
        parts = property_name.split(".")
        return getattr(get_property(index, experiment, ".".join(parts[:-1])), parts[-1])

    properties = ["id", "*"] \
        + sorted(map(lambda s: "config.{s}".format(s=s), experiment.config.parameters.keys())) \
        + sorted(map(lambda s: "parameter.{s}".format(s=s), experiment.parameters.parameters.keys())) \
        + sorted(map(lambda s: "result.{s}".format(s=s), experiment.result.parameters.keys())) \
        + sorted(map(lambda s: "derived.{s}".format(s=s), experiment.derived_callbacks.keys()))

    raise ValueError("Could not find property: {}, choose a valid property ({}) or a filter"
                     .format(property_name, properties))


def is_excluded_from_string(filter_string, experiment):
    if filter_string.startswith("~"):
        return not is_excluded_from_string(filter_string[1:], experiment)

    if "<=" in filter_string:
        parts = filter_string.split("<=", 2)
        value = float(parts[1])
        return get_property(0, experiment, parts[0]) <= value
    if ">=" in filter_string:
        parts = filter_string.split(">=", 2)
        value = float(parts[1])
        return get_property(0, experiment, parts[0]) >= value
    if "<" in filter_string:
        parts = filter_string.split("<", 2)
        value = float(parts[1])
        return get_property(0, experiment, parts[0]) < value
    if ">" in filter_string:
        parts = filter_string.split(">", 2)
        value = float(parts[1])
        return get_property(0, experiment, parts[0]) > value
    if "!=" in filter_string:
        parts = filter_string.split("!=", 2)
        value = float(parts[1])
        return get_property(0, experiment, parts[0]) != value
    if "=" in filter_string:
        parts = filter_string.split("=", 2)
        value = float(parts[1])
        return get_property(0, experiment, parts[0]) == value

    return bool(get_property(0, experiment, filter_string))


def is_excluded(experiment: Experiment, exclude: Optional[List]):
    if exclude is None:
        exclude = []

    if not isinstance(exclude, collections.Iterable):
        exclude = [exclude]

    for f in exclude:
        if callable(f):
            if f(experiment):
                return True
        else:
            if is_excluded_from_string(str(f), experiment):
                return True
    return False


def show(
        experiments: List[Experiment],
        targets=None,
        group_by=None,
        aggregator=None,
        sort: Optional[str] = None,
        exclude: Optional[List] = None,
        plot=None,
        options=None,
        export_filename=None,
):
    # Setup aggregator
    if aggregator == "count":
        aggregator = len
    elif aggregator == "mean" or aggregator is None:
        aggregator = mean
    else:
        raise RuntimeError("Unknown aggregator {}".format(aggregator))

    experiments = [e for e in experiments if not is_excluded(e, exclude)]

    if plot and len(targets) != 1:
        raise ValueError("Plotting requires exactly one target, {} given".format(len(targets)))

    group_by = group_by or (["id"] if plot else ["*"])

    if plot:
        targets = group_by[:1] + targets
        group_by = group_by[1:]

    if sort is not None:
        sort = sort.strip()
        if sort.startswith("-"):
            reverse = True
            sort = sort[1:]
        else:
            reverse = False

        experiments = [
            t[1] for t in sorted(zip(range(len(experiments)), experiments),
                                 key=lambda t: get_property(t[0], t[1], sort),
                                 reverse=reverse)
        ]

    groups = {}
    for i, experiment in enumerate(experiments):
        key = tuple([get_property(i, experiment, p) for p in group_by])
        if key not in groups:
            groups[key] = []
        groups[key].append(tuple([get_property(i, experiment, t) for t in targets]))

    keys = sorted(groups.keys())
    key_names = [["{}:{}".format(g, v) for g, v in zip(group_by, values)] for values in keys]

    if plot:
        from .plot import ScatterData
        scatter = ScatterData("", options)
        for n, k in zip(key_names, keys):
            name = ", ".join(map(str, n))
            sub_group = {}
            for r in groups[k]:
                if r[0] not in sub_group:
                    sub_group[r[0]] = []
                sub_group[r[0]].append(r[1])
            sub_keys = sorted(sub_group.keys())
            scatter.add_data(name, np.array(sub_keys), np.array([aggregator(sub_group[sk]) for sk in sub_keys]),
                             np.array([np.std(np.array(sub_group[sk])) for sk in sub_keys]))

        label_x = targets[0].capitalize()
        label_y = targets[1].capitalize()
        scatter.plot(export_filename,
                     lines=True, log_x=False, log_y=False, label_y=label_y, label_x=label_x, legend_pos="upper center")

    else:
        deviations = dict()
        for k in keys:
            for i, t in enumerate(targets):
                try:
                    deviations[(k, i, t)] = np.std(np.array([r[i] for r in groups[k]]))
                except TypeError:
                    deviations[(k, i, t)] = None
        result_table = [
            ["{:.4f} (+/- {:.4f})".format(aggregator([r[i] for r in groups[k]]), deviations[(k, i, t)] or 0)
             for i, t in enumerate(targets)] for k in keys
        ]
        print(table(result_table, targets, list(zip(*key_names))))



from setuptools import setup, find_packages

# Distribute: python setup.py sdist upload

setup(
    name='autodora',
    version='0.1.8',
    description='Automate experiments and explore your data',
    url='http://github.com/samuelkolb/pydora',
    author='Samuel Kolb',
    author_email='samuel.kolb@me.com',
    license='MIT',
    packages=find_packages(),
    zip_safe=False,
    install_requires=['matplotlib', 'peewee', 'pebble', 'numpy']
)
